make.mosaic.logLik <- function(d) {
  obj <- likelihood.cache(d)
  L <- obj$L
  p <- obj$p
  const <- obj$const
  n <- nrow(d)
  function(m)
    sum(L[cbind(c(1, m+1), c(m, n))]) + const
}

binFunc <- function(x, f, p) {
  g <- function(p) {
    m <- cbind(p^2 + f*p*(1-p),
               2*p*(1-p)*(1-f),
               (1-p)^2 + f*p*(1-p))
    m2 <- pmax(zapsmall(m), 0) # negative freqs not allowed
    sum(x*log(zapsmall(m2/rowSums(m2))+1e-100))
  }
  if ( length(p) == 1 )
    return(c(g(p), p))
  p0 <- min(p)
  p1 <- max(p)
  if ( abs(p0 - p1) < 1e-8 ) {# Don't bother doing anything
    c(g(p0), p0)
  } else {
    ans <- optimize(g, c(p0, p1), maximum=TRUE)
    c(ans$objective, ans$maximum)
  }
}

likelihood.cache <- function(d) {
  N <- rowSums(d)
  freq.p <- d/N
  p <- freq.p[,1] + freq.p[,2]/2
  f <- 1 - freq.p[,2]/(2 * p * (1 - p))
  f[is.nan(f)] <- 1

  n <- nrow(d)
  i <- rep(1:n, n:1)
  j <- sequence(n:1) + i - 1
  ans <- sapply(mapply(seq, i, j), function(k)
                binFunc(d[k,], f[k], p[k]))

  ij <- cbind(i, j)
  ans.L <- ans.p <- matrix(NA, n, n)
  ans.L[ij] <- ans[1,]
  ans.p[ij] <- ans[2,]

  const <- sum(lfactorial(N) - rowSums(lfactorial(d)) +
               d[,2] * log(2*(1 - f + 1e-100)))
  list(L=ans.L, p=ans.p, const=const)
}

